import math
import pytest
import numpy as np
from dtaidistance import dtw

np.set_printoptions(precision=2, linewidth=120)


def test_penalty_cyclicalshift():
    s1 = np.array([0., 1, 2, 1, 0, 1, 2, 1, 0, 1, 2, 1, 0])
    s2 = np.array([2., 1, 0, 1, 2, 1, 0, 1, 2, 1, 0, 1, 2])
    # plt.plot(s1)
    # plt.plot(s2)
    # plt.show(block=True)
    d1 = dtw.distance(s1, s2)
    d2 = dtw.distance(s1, s2, penalty=1)
    assert d1 == pytest.approx(math.sqrt(10))
    assert d2 == pytest.approx(math.sqrt(14))


if __name__ == "__main__":
    test_penalty_cyclicalshift()



# -*- coding: UTF-8 -*-
"""
dtaidistance
~~~~~~~~~~~~

Time series distance methods.

:author: Wannes Meert
:copyright: Copyright 2017 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
import logging


logger = logging.getLogger("be.kuleuven.dtai.distance")


from . import dtw
try:
    from . import dtw_cc
except ImportError:
    # Try to compile automatically
    # try:
    #     import numpy as np
    #     import pyximport
    #     pyximport.install(setup_args={'include_dirs': np.get_include()})
    #     from . import dtw_c
    # except ImportError:
    # logger.warning("\nDTW C variant not available.\n\n" +
    #                "If you want to use the C libraries (not required, depends on cython), " +
    #                "then run `cd {};python3 setup.py build_ext --inplace`.".format(dtaidistance_dir))
    dtw_cc = None

__version__ = "2.2.0"
__author__ = "Wannes Meert"
__copyright__ = "Copyright 2017-2020 KU Leuven, DTAI Research Group"
__license__ = "Apache License, Version 2.0"



# -*- coding: UTF-8 -*-
"""
dtaidistance.dtw_weighted
~~~~~~~~~~~~~~~~~~~~~~~~~

Dynamic Time Warping (DTW) with custom internal distance function.

:author: Wannes Meert
:copyright: Copyright 2018 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

Weights are represented using a tuple (-x3, -x2, -x1, -x0, x0, x1, x2, x3).
The distance d, used in DTW, is multiplied with factor w(d):

.. code-block::

    ^
w(d)|
    |             /
   3|            +
    |           /
    |          /
   1|      +--+
    |     /
   0+----+--------------->
    0   x0 x1 x2 x3     d

The negative and positive values are used to make a distinction between negative
and postive distances. Thus to differentiate between the case that the function
compared with is higher or lower than the target function.

"""
import logging
import math
from collections import defaultdict, deque
import io
import numpy as np

from .dtw import best_path


logger = logging.getLogger("be.kuleuven.dtai.distance")

try:
    from tqdm import tqdm
except ImportError:
    logger.info('tqdm library not available')
    tqdm = None


def warping_paths(s1, s2, weights=None, window=None, **_kwargs):
    """
    Input: s1 and s2 are time series of length n/l1 and m/l2

    :param s1:
    :param s2:
    :param weights: Weights on s1
    :return: DTW similarity m between s1 and s2, warping paths matrix
    """
    # TODO: copy original function in DTW to support all options and integrate weights
    # print('warping_paths')
    l1 = len(s1)
    l2 = len(s2)
    # print('l1', l1, 'l2', l2)
    if window is None:
        window = max(l1, l2)
    else:
        window += 1  # TODO: 0 should be diagonal, this is now 1
    # print('window', window)
    paths = np.full((l1 + 1, l2 + 1), np.inf)
    paths[0, 0] = 0
    for i in range(l1):
        # print('i', i)
        # for j in range(max(0, i - max(0, r - c) - window + 1), min(c, i + max(0, c - r) + window)):
        j_start = max(0, i - max(0, l1 - l2) - window + 1)
        j_end = min(l2, i + max(0, l2 - l1) + window)
        # print(j_start, j_end)
        for j in range(j_start, j_end):
            # print('j', j)
            # for j in range(1, l2 + 1):
            d = s1[i] - s2[j]
            # print(f'd[{i},{j}] = {d}')
            if weights is not None:
                # print(weights[i, :])
                # multiplication with LeRu like function
                xn3, xn2, xn1, xn0, xp0, xp1, xp2, xp3 = weights[i, :]
                # print('xn1, xn0, xp0, xp1', xn1, xn0, xp0, xp1)
                if d < 0:
                    x0, x1, x2, x3 = xn0, xn1, xn2, xn3
                    d = -d
                else:
                    x0, x1, x2, x3 = xp0, xp1, xp2, xp3
                if d <= x0:
                    d = 0
                elif x0 < d < x1:
                    d *= (d - x0) / (x1 - x0)
                elif x2 <= d:
                    if np.isinf(x3) or x3 == x1:
                        a = 1
                    else:
                        a = 2 / (x3 - x2)
                    d *= (1 + a * (d - x2))
                else:
                    pass  # keep d
            # print('d\'', d)
            cost = d ** 2
            paths[i + 1, j + 1] = cost + min(paths[i, j + 1], paths[i + 1, j], paths[i, j])
    # s = math.sqrt(paths[l1 - 1, l2 - 1])
    paths = np.sqrt(paths)
    s = paths[l1 - 1, l2 - 1]
    return s, paths


def distance_matrix(s, weights, window=None, show_progress=False, **kwargs):
    dist_opts = {
        'window': window
    }
    dists = np.full((len(s), len(s)), np.inf)
    it_r = range(len(s))
    if show_progress:
        it_r = tqdm(it_r)
    for r in it_r:
        it_c = range(r + 1, len(s))
        for c in it_c:
            # Because of the weights this is not symmetric (?)
            # TODO: make function not hardcoded
            # print(f'{r} -- {c}')
            # print(f's[{r}]', s[r])
            # print('weights', weights.get(r, None))
            # print(f's[{c}]', s[c])
            # print('weights', weights.get(c, None))
            weights_r = weights.get(r, None)
            d1, paths = warping_paths(s[r], s[c], weights_r, **dist_opts)
            # print(f'd1(r)={d1}  -- w=\n{weights_r}')
            # print (paths)
            weights_c = weights.get(c, None)
            if weights_r is None and weights_c is None:
                dists[r, c] = d1
            else:
                d2, paths = warping_paths(s[c], s[r], weights_c, **dist_opts)
                # print(f'd2(c)={d2}  -- w=\n{weights_c}')
                # print(paths)
                dists[r, c] = min(d1, d2)

    return dists


def compute_weights_using_dt(series, labels, prototypeidx, **kwargs):
    """Compute weight array by learning an ensemble of Decision Trees representing
    the differences between the different labels.

    :param series: List of sequences
    :param labels: Labels for series
    :param prototypeidx: The target sequence to learn weights for
    :param kwargs: Arguments to pass to `series_to_dt`
    :return:
    """
    ml_values, cl_values, _clfs, importances = series_to_dt(series, labels, prototypeidx, **kwargs)
    weights = compute_weights_from_mlclvalues(series[prototypeidx], ml_values, cl_values, **kwargs)
    return weights, importances


def series_to_dt(series, labels, prototypeidx, classifier=None, max_clfs=None, min_ig=0,
                 savefig=None, warping_paths_fnc=None, ignore_idxs=None, patternlen=None,
                 min_purity=1.0, **kwargs):
    """Compute Decision Tree from series

    :param series:
    :param labels: 0 for cannot-link, 1 for must-link
    :param prototypeidx:
    :param classifier: Classifier class.
        For example dtw_weighted.DecisionTreeClassifier or tree.DecisionTreeClassifier
    :param max_clfs: Maximum number of classifiers to learn
    :param min_ig: Minimum information gain
    :param savefig: Path to filename to save tree Graphviz visualisation
    :param warping_paths_fnc: Function to compute warping paths
    :param patternlen: Pattern window size (default None)
    :param kwargs: Passed to warping_paths_fnc
    :return:
    """
    if warping_paths_fnc is None:
        warping_paths_fnc = warping_paths
    if ignore_idxs is None:
        ignore_idxs = set()
    features = [[0] * (len(series[prototypeidx]) * 2)]  # feature per idx, split in positive and negative
    targets = [0]  # Do cluster
    ml_values = defaultdict(lambda: ([], []))

    # print(f"prototype idx = {prototypeidx}")
    for idx, label in enumerate(labels):
        if idx in ignore_idxs:
            continue
        cur_features = np.zeros(len(series[prototypeidx]) * 2, dtype=np.double)
        cur_features_cnt = np.zeros(len(series[prototypeidx]) * 2, dtype=np.int)
        wp_params = {key: kwargs[key] for key in {'window', 'max_dist', 'max_step', 'max_length_diff',
                     'penalty', 'psi'}.intersection(kwargs.keys())}
        s, paths = warping_paths_fnc(series[prototypeidx], series[idx], **wp_params)
        path = best_path(paths)
        for i_to, i_from in path:
            d = series[prototypeidx][i_to] - series[idx][i_from]
            # print(f"d{idx}({i_to},{i_from}) = {d}")
            if label == 0:
                # Cannot-link
                pass
            elif label == 1:
                # Must-link
                if d < 0:
                    ml_values[i_to][0].append(-d)
                elif d > 0:
                    ml_values[i_to][1].append(d)
            if d <= 0:
                cur_features[i_to * 2] += -d
                cur_features_cnt[i_to * 2] += 1
            if d >= 0:
                cur_features[i_to * 2 + 1] += d
                cur_features_cnt[i_to * 2 + 1] += 1

        cur_features_cnt[cur_features_cnt == 0] = 1
        cur_features = np.divide(cur_features, cur_features_cnt)
        features.append(cur_features)
        if label == 0:
            # print(f"{idx:<2}: CL -> {cur_features[0]:0.3f} / {cur_features[1]:0.3f}")
            targets.append(1)  # Do not cluster
        elif label == 1:
            # print(f"{idx:<2}: ML -> {cur_features[0]:0.3f} / {cur_features[1]:0.3f}")
            targets.append(0)  # Do cluster
        else:
            raise Exception("Encountered a label that is not 0 (cannot-link) or 1 (must-link): {}".format(label))

    if classifier is None:
        classifier = DecisionTreeClassifier

    features = np.array(features)
    targets = np.array(targets)

    if savefig is not None:
        try:
            from sklearn import tree
        except ImportError:
            logger.error("No figure generated, sklearn is not installed.")
            savefig, tree, out_string, feature_names = None, None, None, None
        out_string = io.StringIO()

        def args(i):
            if (i % 2) == 0:
                sgn = '-'
                cmp = 's>t'
            else:
                sgn = '+'
                cmp = 's<t'
            return i, sgn, cmp

        feature_names = ["d[{}] ({}, {}, {})".format(i // 2, *args(i))
                         for i in range(2*len(series[prototypeidx]))]
        class_names = ["ML", "CL"]
    else:
        tree, out_string, feature_names, class_names = None, None, None, None

    if patternlen is not None:
        cl_values, clfs, importances = dt_windows(features, targets, classifier, patternlen, max_clfs, min_ig, min_purity)
    else:
        cl_values, clfs, importances = dt_onewindow(features, targets, classifier, max_clfs, min_ig, min_purity)

    if savefig is not None:
        for clf in clfs:
            tree.export_graphviz(clf, out_file=out_string, feature_names=feature_names, class_names=class_names)
            print("\n\n", file=out_string)
        with open(savefig, "w") as ofile:
            print(out_string.getvalue(), file=ofile)

    return ml_values, cl_values, clfs, importances


def dt_windows(features, targets, classifier, patternlen, max_clfs, min_ig, min_purity):
    cl_values = dict()
    clfss = []
    ignore_features = set()
    clf_w = 1.0
    importances = defaultdict(lambda: [0, 0])

    nb_features = features.shape[1]
    nb_indices = int((nb_features / 2))
    max_kd = 0
    for idx in range(0, int(nb_indices - patternlen / 2), int(patternlen / 2)):
        idx_s = idx * 2
        idx_e = idx_s + patternlen * 2
        clf = classifier()
        cur_features = features[:, idx_s:idx_e]
        clf.fit(cur_features, targets, ignore_features=ignore_features, min_ig=min_ig)
        logger.debug(f"Learned classifier {len(clfss) + 1}: idx = f{idx}/{idx_s}:f{idx+patternlen}/{idx_e}, "
                     f"nb nodes = {clf.tree_.nb_nodes}, used features = {clf.tree_.used_features}")
        if clf.tree_.nb_nodes <= 1:
            continue
        clf.set_features(list(range(idx_s, idx_e)))
        max_kd = max(max_kd, np.max(clf.tree_.kd))
        clfss.append(clf)
    clfs = []
    for clf_idx, clf in enumerate(clfss):
        score = clf.score(max_kd)
        logger.debug(f"Clf[{clf_idx:<2}] - Score = {score}, Entropy = {clf.avg_impurity()}, "
                     f"depth = {clf.tree_.depth}, nbnodes = {clf.tree_.nb_nodes}")
        clfs.append((score, -clf.tree_.nb_nodes, clf))
    clfs.sort(reverse=True)
    min_score = clfs[-1][0]
    max_score = clfs[0][0]
    minallowed_score = max_score - (max_score - min_score) / 5  # TODO: remove this magic number
    clfs_use = 1
    # We expect that the current range includes at least two new classifiers
    while len(clfs) > clfs_use and minallowed_score > clfs[clfs_use][0]:
        max_score = clfs[clfs_use][0]
        minallowed_score = max_score - (max_score - min_score) / 5
        clfs_use += 1
    clfs = [t for t in clfs if t[0] >= minallowed_score]

    if max_clfs is not None:
        clfs = clfs[:max_clfs]
    logger.debug(f"Kept {len(clfs)} classifiers with score >= {minallowed_score}, clfs_use = {clfs_use}")
    for clf_score, clf_nbnodes, clf in clfs:
        new_cl_values, used_features = decisiontree_to_clweights(clf, min_purity)
        # if len(used_features) == 0:
        #     logger.debug(f"No features used, ignore all features in tree: {clf.tree_.used_features}")
        #     used_features.update(clf.tree_.used_features)
        update_cl_values(cl_values, new_cl_values)
        update_importances(importances, new_cl_values, (clf_score - min_score) / (max_score - min_score))
        # print(f"new_cl_values: {new_cl_values}")
        # print(f"cl_values: {cl_values}")
        # ignore_features.update(clf.tree_.used_features)
        ignore_features.update(used_features)
        # print(f"ignore_features: {ignore_features}")
    return cl_values, clfs, importances


def dt_onewindow(features, targets, classifier, max_clfs, min_ig, min_purity):
    cl_values = dict()
    clfs = []
    ignore_features = set()
    not_empty = True
    clf_w = 1.0
    importances = defaultdict(lambda: [0, 0])
    while not_empty and not (max_clfs is not None and len(clfs) >= max_clfs):
        clf = classifier()
        clf.fit(features, targets, ignore_features=ignore_features, min_ig=min_ig)
        logger.debug(f"Learned classifier {len(clfs) + 1}: nb nodes = {clf.tree_.nb_nodes}")
        if clf.tree_.nb_nodes <= 1:
            not_empty = False
            continue
        clfs.append(clf)
        new_cl_values, used_features = decisiontree_to_clweights(clf, min_purity)
        if len(used_features) == 0:
            logger.debug(f"No features used, ignore all features in tree: {clf.tree_.used_features}")
            used_features.update(clf.tree_.used_features)
        update_cl_values(cl_values, new_cl_values)
        update_importances(importances, new_cl_values, clf_w)
        # print(f"new_cl_values: {new_cl_values}")
        # print(f"cl_values: {cl_values}")
        # ignore_features.update(clf.tree_.used_features)
        ignore_features.update(used_features)
        # print(f"ignore_features: {ignore_features}")
        clf_w *= 0.66
    return cl_values, clfs, importances


def update_cl_values(cl_values, new_cl_values):
    for idx, (n, p) in new_cl_values.items():
        if idx not in cl_values:
            cl_values[idx] = [n, p]
        else:
            cl_values[idx][0].extend(n)
            cl_values[idx][1].extend(p)


def update_importances(importances, new_cl_values, weight):
    for idx, (n, p) in new_cl_values.items():
        if len(n) > 0:
            importances[idx][0] = max(weight, importances[idx][0])
        if len(p) > 0:
            importances[idx][1] = max(weight, importances[idx][1])


def decisiontree_to_clweights(clf, min_purity=1.0):
    """Translate a decision tree to a set of cannot-link weights.

    This is based on the concept that we can represent the false (class 1) case as the
    conjunction of all negated false (class 1) leafs. It will be certainly not class 1
    if it is "not any of the leafs that say class 1" and thus "!(leaf_1 v leaf_2 v ... v leaf_n)"
    which is "!leaf_1 ^ !leaf_2 ^ ... ^ !leaf_n".

    :param clf: DecisionTreeClassifier
    :returns: Weights
    """
    dtnodes = deque([(0, [])])
    cl_values = defaultdict(lambda: ([], []))
    used_features = set()

    while len(dtnodes) > 0:
        curnode, path = dtnodes.popleft()
        if clf.tree_.children_left[curnode] == -1 and clf.tree_.children_right[curnode] == -1:
            value = clf.tree_.value[curnode][0]
            # logger.debug(f"Leaf - values = {value}")
            purity = value[1] / (value[0] + value[1])
            # print(f"purity: {purity}, values: {value}")
            assert((value[0] == 0 and purity == 1.0) or purity != 1.0)
            if purity >= min_purity:
                # Leaf that represents cannot-link with given purity
                # logger.debug(f'CL leaf: {curnode}')
                cur_used_features = clweights_updatefrompath(cl_values, path)
                used_features.update(cur_used_features)
        else:
            threshold = clf.tree_.threshold[curnode]
            feature = clf.tree_.feature[curnode]
            path_left = path + [(feature, threshold, True)]  # true branch (f <= t)
            dtnodes.append((clf.tree_.children_left[curnode], path_left))
            path_right = path + [(feature, threshold, False)]  # false branch (f > t)
            dtnodes.append((clf.tree_.children_right[curnode], path_right))
    return cl_values, used_features


def clweights_updatefrompath(cl_values, path):
    logger.debug(f"Path to CL: {path}")
    used_features = set()
    for feature, threshold, leq in path:
        index = feature // 2
        dneg = ((feature % 2) == 0)
        if leq:  # f <= t
            logger.debug(f"Ignore: CL with f{index} <= {threshold} (d is negative={dneg}, feature={feature})")
        else:  # f > t
            logger.debug(f"Accept: CL with f{index} >  {threshold} (d is negative={dneg}, feature={feature})")
            cl_values[index][0 if dneg else 1].append(threshold)
            used_features.add(feature)
    return used_features


def compute_weights_from_mlclvalues(serie, ml_values, cl_values, only_max=False, strict_cl=True, **_kwargs):
    """Compute weights that represent a rectifier type of function based on must-link values and cannot-link values.

         |             /
        3|            +
         |           /
         |          /
        1|      +--+
         |     /
        0+----+
         0   x0 x1 x2 x3

    """
    if __debug__ and logger.isEnabledFor(logging.DEBUG):
        logger.debug('ml_values = ' + ('None' if ml_values is None else str(list(ml_values.items()))))
        logger.debug('cl_values = ' + ('None' if cl_values is None else str(list(cl_values.items()))))

    wn = np.zeros((len(serie), 8), dtype=np.double)  # xns, xn2, xn1, xn0, xp0, xp1, xp2, xps
    wn[:, 0:2] = np.inf
    wn[:, 6:8] = np.inf

    # First find ml-max and cl-min values
    ml_maxmin_n = np.zeros((len(serie), 3))
    ml_maxmin_p = np.zeros((len(serie), 3))
    for idx in range(len(serie)):
        if idx in ml_values:
            mls = ml_values[idx][0]
        else:
            mls = []
        if idx in cl_values:
            cls = cl_values[idx][0]
        else:
            cls = []
        ml_max = _clean_max(mls, cls)
        cl_min = _clean_min(cls, mls, strict_cl)
        if np.isinf(cl_min):
            diff = 0
        else:
            diff = cl_min - ml_max
        ml_maxmin_n[idx, :] = [ml_max, cl_min, diff]

        if idx in ml_values:
            mls = ml_values[idx][1]
        else:
            mls = []
        if idx in cl_values:
            cls = cl_values[idx][1]
        else:
            cls = []
        ml_max = _clean_max(mls, cls)
        cl_min = _clean_min(cls, mls, strict_cl)
        if np.isinf(cl_min):
            diff = 0
        else:
            diff = cl_min - ml_max
        ml_maxmin_p[idx, :] = [ml_max, cl_min, diff]

    if only_max:
        # Only retain the cl-min value where the distance between ml_max and cl_min is largest
        maxval = np.max(ml_maxmin_n[:, 2])
        maxidx = ml_maxmin_n[:, 2] == maxval
        # print('maxval', maxval)
        # print('maxidx', maxidx)
        # print('values', ml_maxmin_n)
        vals = ml_maxmin_n[:, 1][maxidx]
        ml_maxmin_n[:, 1] = np.inf
        ml_maxmin_n[:, 1][maxidx] = vals

        maxval = np.max(ml_maxmin_p[:, 2])
        maxidx = ml_maxmin_p[:, 2] == maxval
        # print('maxval', maxval)
        # print('maxidx', maxidx)
        # print('values', ml_maxmin_p)
        vals = ml_maxmin_p[:, 1][maxidx]
        ml_maxmin_p[:, 1] = np.inf
        ml_maxmin_p[:, 1][maxidx] = vals

    for idx in range(len(serie)):
        # vn3, vn2, vn1, vn0, vp0, vp1, vp2, vp3 = wn[idx]
        # Negative
        vn1 = 1.5 * ml_maxmin_n[idx, 0]
        vn3 = ml_maxmin_n[idx, 1]
        if vn1 > vn3:
            vn1 = vn3
        vn0 = 0.5 * vn1
        vn2 = 0.9 * vn3
        if vn2 < vn1:
            vn1 = vn2 = (vn1 + vn2) / 2
        # Positive
        vp1 = 1.5 * ml_maxmin_p[idx, 0]
        vp3 = ml_maxmin_p[idx, 1]
        if vp1 > vp3:
            vp1 = vp3
        vp0 = 0.5 * vp1
        vp2 = 0.9 * vp3
        if vp2 < vp1:
            vp1 = vp2 = (vp1 + vp2) / 2
        wn[idx, :] = [vn3, vn2, vn1, vn0, vp0, vp1, vp2, vp3]
    # logger.debug(f'weights:\n{wn}')
    return wn


def _clean_max(mls, cls):
    """Return the maximal value of mls that is smaller than all values in cls."""
    mls.sort()
    cls.sort()
    if cls is None or len(cls) == 0:
        min_cls = np.inf
    else:
        min_cls = cls[0]
    max_mls = 0
    for ml in mls:
        if ml > min_cls:
            return max_mls
        if ml > max_mls:
            max_mls = ml
    return max_mls


def _clean_min(cls, mls, keep_largest=True):
    """Return the minimal value of cls that is larger than all values in mls."""
    mls.sort()
    cls.sort()
    min_cls = np.inf
    if mls is None or len(mls) == 0:
        max_mls = 0
    else:
        max_mls = mls[-1]
    for cl in reversed(cls):
        if cl < max_mls:
            break
        if cl < min_cls:
            min_cls = cl
    if keep_largest:
        # If None of the cl values are kept, use the largest value.
        if np.isinf(min_cls) and cls is not None and len(cls) > 0:
            min_cls = cls[-1]
    return min_cls


def plot_margins(serie, weights, filename=None, ax=None, origin=(0, 0), scaling=(1, 1), y_limit=None,
                 importances=None):
    try:
        from matplotlib import pyplot as plt
    except ImportError:
        print("Matplotlib not installed")
        return
    if weights is None:
        return
    if y_limit is None:
        y_limit = (np.min(serie), np.max(serie))
        diff = y_limit[1] - y_limit[0]
        y_limit = (y_limit[0] - diff * 0.2, y_limit[1] + diff * 0.2)
    # logger.debug(f"y_limit = {y_limit}")
    if ax is None:
        fig, ax = plt.subplots(nrows=1, ncols=1)
    else:
        fig = None

    # logger.debug(f"Weights =\n{weights}")

    #  |             /
    # 3|            +
    #  |           /
    #  |          /
    # 1|      +--+
    #  |     /
    # 0+----+
    #  0   x0 x1 x2 x3
    xn3 = weights[:, 0]
    xn2 = weights[:, 1]
    xn1 = weights[:, 2]
    xn0 = weights[:, 3]
    xp0 = weights[:, 4]
    xp1 = weights[:, 5]
    xp2 = weights[:, 6]
    xp3 = weights[:, 7]

    # Green, between 0 and xn0
    upper_bound_0 = [origin[1] + scaling[1] * (y + w) for y, w in zip(serie, xn0)]
    lower_bound_0 = [origin[1] + scaling[1] * (y - w) for y, w in zip(serie, xp0)]
    ax.fill_between(range(len(serie)), lower_bound_0, upper_bound_0, facecolor='green', alpha=0.2)

    # Dark Green, between x0 and x1
    upper_bound_1 = np.array([origin[1] + scaling[1] * (y + w) for y, w in zip(serie, xn1)])
    lower_bound_1 = np.array([origin[1] + scaling[1] * (y - w) for y, w in zip(serie, xp1)])
    lower_bound_1[lower_bound_1 < (origin[1] + y_limit[0])] = origin[1] + y_limit[0]
    upper_bound_1[upper_bound_1 > (origin[1] + y_limit[1])] = origin[1] + y_limit[1]
    ax.fill_between(range(len(serie)), lower_bound_0, lower_bound_1, facecolor='green', alpha=0.1)
    ax.fill_between(range(len(serie)), upper_bound_0, upper_bound_1, facecolor='green', alpha=0.1)

    # Cyan , between x1 and x2
    # xn2[np.isinf(xn2)] = abs(y_limit[0])
    # xp2[np.isinf(xp2)] = abs(y_limit[1])
    upper_bound_2 = np.array([origin[1] + scaling[1] * (y + w) for y, w in zip(serie, xn2)])
    lower_bound_2 = np.array([origin[1] + scaling[1] * (y - w) for y, w in zip(serie, xp2)])
    upper_bound_2[upper_bound_2 > (origin[1] + y_limit[1])] = origin[1] + y_limit[1]
    lower_bound_2[lower_bound_2 < (origin[1] + y_limit[0])] = origin[1] + y_limit[0]
    # ax.fill_between(range(len(serie)), lower_bound_1, lower_bound_2, facecolor='cyan', alpha=0.1)
    # ax.fill_between(range(len(serie)), upper_bound_1, upper_bound_2, facecolor='cyan', alpha=0.1)

    # Red, between x2 and x3
    # xn3[np.isinf(xn3)] = abs(y_limit[1])
    # xp3[np.isinf(xp3)] = abs(y_limit[0])
    upper_bound_3 = np.array([origin[1] + scaling[1] * (y + w) for y, w in zip(serie, xn3)])
    lower_bound_3 = np.array([origin[1] + scaling[1] * (y - w) for y, w in zip(serie, xp3)])
    upper_bound_3[upper_bound_3 > (origin[1] + y_limit[1])] = origin[1] + y_limit[1]
    lower_bound_3[lower_bound_3 < (origin[1] + y_limit[0])] = origin[1] + y_limit[0]
    ax.fill_between(range(len(serie)), lower_bound_2, lower_bound_3, facecolor='red', alpha=0.1)
    ax.fill_between(range(len(serie)), upper_bound_2, upper_bound_3, facecolor='red', alpha=0.1)

    # Dark Red, between x3 and limit
    lower_bound_4 = np.array([origin[1] + y_limit[0] for y in serie])
    upper_bound_4 = np.array([origin[1] + y_limit[1] for y in serie])
    ax.fill_between(range(len(serie)), lower_bound_3, lower_bound_4, facecolor='red', alpha=0.2)
    ax.fill_between(range(len(serie)), upper_bound_3, upper_bound_4, facecolor='red', alpha=0.2)

    # Importances
    if importances is not None:
        for idx, (imp_n, imp_p) in importances.items():
            if imp_n != 0:
                y = origin[1] + y_limit[1]
                alpha = (imp_n + 0.2) / 1.2
                ax.plot([idx], [y], marker='o', color='red', alpha=alpha)
            if imp_p != 0:
                y = origin[1] + y_limit[0]
                alpha = (imp_p + 0.2) / 1.2
                ax.plot([idx], [y], marker='o', color='red', alpha=alpha)

    # Plot series
    ax.plot(origin[1] + scaling[1] * serie)

    if filename:
        plt.savefig(filename, bbox_inches='tight', pad_inches=0)
        plt.close()
        fig, ax = None, None
    return fig, ax


class DecisionTreeClassifier:
    def __init__(self):
        """Decision Tree that uses a combination of Information Gain and k-distance.
        It also offers the option to use every feature only once.


        Note:
        - This classifier assumes only two classes: 0 and 1.
        """
        self.tree_ = None
        self.criterion = "igkdistance"
        self.n_features_ = None

    @staticmethod
    def entropy(targets):
        """H(X) = -\\sum_{i=0}^{n}(P(x_i))\\log P(x_i))

        :param targets: Numpy array with class target (values are 0 and 1)
        :return: entropy
        """
        # print(f'entropy({targets})')
        n = len(targets)
        if n == 0:
            raise Exception("Empty list passed to entropy")
        prob = np.sum(targets) / n
        if prob == 0 or prob == 1:
            h = 0
        else:
            h = - prob * math.log(prob) - (1.0 - prob) * math.log(1.0 - prob)
        # print(f'entropy - targets: {targets} - h: {h}')
        return h

    @staticmethod
    def informationgain_continuous(targets, values, h0=None):
        """Best split based on information gain.

        :return: (information gain, threshold value, entropy of current node)
        """
        # print(f'entropy_continuous:\n{targets}\n{values}')
        if h0 is None:
            h0 = DecisionTreeClassifier.entropy(targets)
        # print(f'h0={h0}, targets={targets}')

        thresholds = np.unique(values)
        thresholds = (thresholds[1:] + thresholds[:-1]) / 2
        n = len(values)
        h1_min, th_min = np.inf, None
        for threshold in thresholds:
            mask = (values <= threshold)
            prob = np.sum(mask) / n
            h1 = prob * DecisionTreeClassifier.entropy(targets[mask]) + \
                 (1.0 - prob) * DecisionTreeClassifier.entropy(targets[~mask])
            # print(f'th={threshold} - h1={h1} - prob={prob} - ig={h0 - h1}')
            if h1 < h1_min:
                h1_min = h1
                th_min = threshold
                # No need to do IGR, it always splits in two
                # ig = h0 - h1
                # if prob == 0.0 or prob == 1.0:
                #     iv = 0
                # else:
                #     iv = -prob * math.log(prob, 2) - (1.0 - prob) * math.log(1.0 - prob, 2)
                # if iv > iv_max:
                #     iv_max = iv
                #     th_max = threshold
        if th_min is None:
            ig = 0
        else:
            ig = h0 - h1_min
        return ig, th_min, h0

    @staticmethod
    def kdistance(values, threshold, k=5):
        """k-distance density measure .

        :param values:
        :param threshold:
        :param k: k-th element to return distance of
        :return: Distances to k nearest neighbours
        """
        # print(f'kdistance({values}, {threshold})')
        dists = []
        for value in np.nditer(values):
            dist = abs(value - threshold)
            if len(dists) < k:
                dists.append(dist)
                dists.sort()
            elif dist < dists[-1]:
                dists[-1] = dist
                dists.sort()
        # Sometimes k-distance is defined as:
        # dk = 1/k*sum([d**2 for d in dists])
        return dists[-1]

    def fit(self, features, targets, use_feature_once=True, ignore_features=None, min_ig=0):
        """Learn decision tree.

        :param features: Array of dimension #instances x #features, type float
        :param targets: Array of dimension #instances, values 0 or 1
        :param use_feature_once: Use each feature only once in a path
        :param ignore_features: Set of feature indices to ignore
        :param min_ig: Minimal information gain
        :return: DecisionTreeClassifier object (self)

        Fills the tree_ variable with a decision tree representation.
        """
        # print(f'features:\n{features}')
        # print(f'targets:\n{targets}')
        nb_features = features.shape[1]
        self.n_features_ = nb_features
        nb_instances = features.shape[0]
        # print(f'nb_instances: {nb_instances} (targets.shape = {targets.shape}')
        k = int(math.ceil(len(targets) * 0.005))
        k_relative = float(np.max(features))
        self.tree_ = Tree()
        queue = deque([(self.tree_.last(),                  # Leaf
                        np.zeros(nb_features, dtype=bool),  # Used features
                        np.ones(nb_instances, dtype=bool)   # Instances mask
                        )])
        queue_it = 0
        while len(queue) > 0:
            queue_it += 1
            node, used_ftrs, idxs = queue.popleft()
            # print(f'------ node ({queue_it})\n  {node}\n  {used_ftrs}\n  {idxs}')
            nb_samples = np.sum(idxs)
            targetsum = np.sum(targets[idxs])
            self.tree_.value[node][0, 1] = targetsum
            nontargetsum = nb_samples - targetsum
            self.tree_.value[node][0, 0] = nontargetsum
            self.tree_.n_node_samples[node] = nb_samples
            if np.all(targets[idxs]) or not np.any(targets[idxs]):
                # print('Pure leaf')
                self.tree_.impurity[node] = 0
                continue
            curvalues = features[idxs, :]
            curtargets = targets[idxs]
            h0 = DecisionTreeClassifier.entropy(curtargets)
            all_gains = []
            max_kd = 0
            for fi in range(nb_features):
                if (use_feature_once and used_ftrs[fi]) or (ignore_features is not None and fi in ignore_features):
                    continue
                ig, thr, _ = self.informationgain_continuous(curtargets, curvalues[:, fi], h0=h0)
                if thr is None or ig < min_ig:
                    kd = None
                    gain = 0.0
                    # if __debug__ and logger.isEnabledFor(logging.DEBUG):
                    #     if thr is None:
                    #         thr = "None"
                    #     else:
                    #         thr = f"{thr:+.5f}"
                    #     logger.debug(f"Not splitting feature {fi:<3}, ig={ig:.5f}, thr={thr}, "
                    #                  f"k={k}, kd=None, gain=0.0")
                else:
                    # Prefer values in low-density regions (thus large k dist)
                    kd = self.kdistance(curvalues[:, fi], thr, k=k)
                    if kd > max_kd:
                        max_kd = kd
                    all_gains.append((fi, ig, thr, kd))
            best_gain, best_fi, best_thr, best_kd = 0, None, None, None
            for fi, ig, thr, kd in all_gains:
                gain = ig * (1 + (kd / max_kd)**1)
                # if __debug__ and logger.isEnabledFor(logging.DEBUG):
                #     logger.debug(f"Splitting feature {fi:<3}, ig={ig:.5f}, thr={thr:+.5f}, "
                #                  f"k={k}, kd={kd:.5f}, gain={gain:.5f}")
                # print(f'fi={fi}, thr={thr}, ig={ig}, kd={kd}, gain={gain}')
                if best_gain < gain:
                    best_gain = gain
                    best_fi = fi
                    best_thr = thr
                    best_kd = kd
            self.tree_.impurity[node] = h0
            if best_fi is not None:
                self.tree_.kd[node] = best_kd
                leq_idxs = idxs & (features[:, best_fi] <= best_thr)
                used_ftrs = used_ftrs.copy()
                used_ftrs[best_fi] = True
                self.tree_.feature[node] = best_fi
                self.tree_.threshold[node] = best_thr
                lessorequal = self.tree_.add()
                # print(f'best_fi={best_fi}, best_thr={best_thr}, best_gain={best_gain}')
                queue.append((lessorequal, used_ftrs, leq_idxs))
                self.tree_.children_left[node] = lessorequal
                larger = self.tree_.add()
                larger_idxs = idxs & (features[:, best_fi] > best_thr)
                queue.append((larger, used_ftrs, larger_idxs))
                self.tree_.children_right[node] = larger
                # print('New queue:')
                # print('\n'.join([str(elmt) for elmt in queue]))
            else:
                pass
        return self

    def avg_entropy_leafs(self):
        e = 0
        c = 0
        for i in range(len(self.tree_.feature)):
            if self.tree_.feature[i] == -1:
                c += 1
                values = self.tree_.value[i][0]
                prob = values[0] / (values[0] + values[1])
                try:
                    e -= prob * math.log(prob) + (1-prob) * math.log(1-prob)
                except ValueError:
                    pass
        return e / c

    def avg_impurity(self):
        impurity = 0
        count = 0
        for idx in range(len(self.tree_.feature)):
            if self.tree_.feature[idx] == -1:
                count += 1
                impurity += self.tree_.impurity[idx]
        impurity = impurity / count
        return impurity

    def score(self, max_kd):
        # max_kd = np.max(self.tree_.kd)
        sum_wimpurity = 0
        cnt_wimpurity = 0
        queue = deque()
        queue.append((0, 0, 0))  # index, kd max, kd count
        while len(queue) > 0:
            idx, branch_max_kd, cnt = queue.popleft()
            idx_l = self.tree_.children_left[idx]
            idx_r = self.tree_.children_right[idx]
            if idx_l == -1 and idx_r == -1:
                # print(f"impurity = {self.tree_.impurity[idx]}, branch_max_kd = {branch_max_kd}, max_kd = {max_kd}, factor = {1 + (branch_max_kd / max_kd)**2}")
                sum_wimpurity += (1.0 - self.tree_.impurity[idx]) * (1 + (branch_max_kd / max_kd)**1)
                cnt_wimpurity += 1
            else:
                ncsum = max(branch_max_kd, self.tree_.kd[idx])
                ncnt = cnt + 1
                queue.append((idx_l, ncsum, ncnt))
                queue.append((idx_r, ncsum, ncnt))
        return sum_wimpurity / cnt_wimpurity

    def set_features(self, features):
        for i in range(len(self.tree_.feature)):
            if self.tree_.feature[i] != -1:
                self.tree_.feature[i] = features[self.tree_.feature[i]]

    def __eq__(self, other):
        return True

    def __lt__(self, other):
        return False

    def __le__(self, other):
        return True

    def __gt__(self, other):
        return False

    def __ge__(self, other):
        return True


class Tree:
    def __init__(self):
        """Tree to represent a Decision Tree.

        This datastructure has the fields required to be compatible with scikit-learn and
        its Graphvis DOT representation.
        """
        self.threshold = []
        self.feature = []
        self.children_right = []
        self.children_left = []
        self.n_outputs = 1
        self.value = []
        self.impurity = []
        self.n_node_samples = []
        self.n_classes = []
        self.kd = []
        self.add()

    def add(self):
        """Add a new node to the tree.

        :return: Index of new node
        """
        self.threshold.append(-1)
        self.feature.append(-1)
        self.children_right.append(-1)
        self.children_left.append(-1)
        self.value.append(np.array([[0, 0]]))
        self.impurity.append(-1)
        self.n_node_samples.append(-1)
        self.n_classes.append(-1)
        self.kd.append(-1)
        return len(self.feature) - 1

    def last(self):
        """Last node.

        :return: Index of last added node
        """
        return len(self.feature) - 1

    @property
    def nb_nodes(self):
        return len(self.threshold)

    @property
    def used_features(self):
        return set(self.feature)

    @property
    def depth(self):
        max_depth = 1
        queue = deque()
        queue.append((0, 1))
        while len(queue) > 0:
            idx, depth = queue.popleft()
            if depth > max_depth:
                max_depth = depth
            idx_l = self.children_left[idx]
            idx_r = self.children_right[idx]
            if idx_l != -1:
                queue.append((idx_l, depth + 1))
            if idx_r != -1:
                queue.append((idx_r, depth + 1))
        return max_depth



#!/usr/bin/env python3
# encoding: utf-8
"""
hierarchical_clustering - Example

__author__ = "Wannes Meert"
__copyright__ = "Copyright 2017 KU Leuven, DTAI Research Group"
__license__ = "APL"

..
    Part of the DTAI distance code.

    Copyright 2016 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import sys
import os
import argparse
import logging
import math

import numpy as np
from scipy.cluster.hierarchy import linkage, dendrogram
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D

sys.path.append(os.path.join(os.path.dirname(os.path.realpath(__file__)), os.pardir))
from dtaidistance import dtw

logger = logging.getLogger(__name__)


def size_cond(size):
    n = size
    r = 2
    f = math.factorial
    return int(f(n) / f(r) / f(n-r))


def example_naivehierarchicalclustering():
    """Naive hierarchical clustering algorithm using DTW and based on .

    For a more efficient approach, check:
    Mueen, A and Keogh, E, Extracting Optimal Performance from Dynamic Time Warping,
    Tutorial, KDD 2016
    http://www.cs.unm.edu/~mueen/DTW.pdf

    :return: None
    """
    series = [
        np.array([0.0, 0.0, 2.0, 1.0, 1.0, 0.0, 0.0]),
        np.array([0.0, 0.0, 2.0, 1.0, 1.0, 0.0, 0.0]),
        np.array([2.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0]),
        np.array([2.0, 1.0, 1.0, 0.0, 0.0, 2.0, 3.0]),
        np.array([4.0, 2.0, 1.0, 0.0, 0.0, 1.0, 3.0])
    ]

    dists = dtw.distance_matrix_fast(series)
    print("Distance matrix:\n{}".format(dists))

    dists_cond = np.zeros(size_cond(len(series)))
    idx = 0
    for r in range(len(series)-1):
        dists_cond[idx:idx+len(series)-r-1] = dists[r, r+1:]
        idx += len(series)-r-1

    z = linkage(dists_cond, method='complete', metric='euclidean')

    fig, axes = plt.subplots(2, 1, figsize=(8, 3))
    for idx, serie in enumerate(series):
        serie += idx * 0.1
        axes[0].plot(serie, label=str(idx))
        axes[0].text(0 + 0.15 * (-1)**idx * idx, serie[0] + 0.15 * idx, idx)
        axes[0].add_line(Line2D([0, 0 + 0.15 * (-1)**idx * idx], [serie[0], serie[0] + 0.15 * idx],
                                linewidth=1, color='gray'))
    axes[0].legend(loc=1)
    dendrogram(z, ax=axes[1])
    plt.show(block=True)


def main(argv=None):
    parser = argparse.ArgumentParser(description='Example hierarchical clustering')
    parser.add_argument('--verbose', '-v', action='count', help='Verbose output')
    args = parser.parse_args(argv)

    logger.setLevel(logging.ERROR - 10 * (0 if args.verbose is None else args.verbose))
    logger.addHandler(logging.StreamHandler(sys.stdout))

    example_naivehierarchicalclustering()


if __name__ == "__main__":
    sys.exit(main())



import sys, os
import logging
from pathlib import Path

import pytest

from dtaidistance import dtw, util_numpy
from dtaidistance import dtw_visualisation as dtwvis


numpyonly = pytest.mark.skipif("util_numpy.test_without_numpy()")
logger = logging.getLogger("be.kuleuven.dtai.distance")


def read_twolead(ucr):
    with util_numpy.test_uses_numpy() as np:
        datafile = ucr / "TwoLeadECG" / "TwoLeadECG_TRAIN"
        data = []
        with open(datafile, "r") as ifile:
            for line in ifile:
                data.append([float(v) for v in line.split(",")])
        data = np.array(data)
        series = data[:, 1:]
        labels = data[:, 0]
        return series, labels


@numpyonly
def test_distance_twolead_startdiff(directory=None, ucr=None):
    if ucr is None:
        # skip test
        return
    series, labels = read_twolead(ucr)
    series1 = series[9]
    series2 = series[21]
    # print(labels[9])
    # print(series1)
    # print(labels[21])
    # print(series2)

    window, psi = 10, 2
    d1a = dtw.distance_fast(series1, series2, window=window, psi=psi)
    d1b = dtw.distance(series1, series2, window=window, psi=psi)
    # print(d1a, d1b)

    if directory:
        fn = directory / "warpingpaths_startdiff.png"
        d2, paths = dtw.warping_paths(series1, series2, window=window, psi=psi)
        best_path = dtw.best_path(paths)
        dtwvis.plot_warpingpaths(series1, series2, paths, best_path, filename=fn, shownumbers=False)
        print(f"Figure saved to {fn}")


@numpyonly
def test_distance_twolead_sameclass(directory=None, ucr=None):
    if ucr is None:
        # skip test
        return
    series, labels = read_twolead(ucr)
    series1 = series[9]
    series2 = series[10]
    # print(labels[9])
    # print(series1)
    # print(labels[10])
    # print(series2)

    window, psi = 10, 2
    d1a = dtw.distance_fast(series1, series2, window=window, psi=psi)
    d1b = dtw.distance(series1, series2, window=window, psi=psi)
    # print(d1a, d1b)

    if directory:
        fn = directory / "warpingpaths_sameclass.png"
        d2, paths = dtw.warping_paths(series1, series2, window=window, psi=psi)
        best_path = dtw.best_path(paths)
        dtwvis.plot_warpingpaths(series1, series2, paths, best_path, filename=fn, shownumbers=False)
        print(f"Figure saved to {fn}")


@numpyonly
def test_distance_twolead_diffclass(directory=None, ucr=None):
    if ucr is None:
        # skip test
        return
    series, labels = read_twolead(ucr)
    series1 = series[9]
    series2 = series[11]
    # print(labels[9])
    # print(series1)
    # print(labels[11])
    # print(series2)

    window, psi = 10, 2
    d1a = dtw.distance_fast(series1, series2, window=window, psi=psi)
    d1b = dtw.distance(series1, series2, window=window, psi=psi)
    # print(d1a, d1b)

    if directory:
        fn = directory / "warpingpaths_diffclass.png"
        d2, paths = dtw.warping_paths(series1, series2, window=window, psi=psi)
        best_path = dtw.best_path(paths)
        dtwvis.plot_warpingpaths(series1, series2, paths, best_path, filename=fn, shownumbers=False)
        print(f"Figure saved to {fn}")


@numpyonly
def test_distance_twolead_matrix(directory=None, ucr=None):
    if ucr is None:
        # skip test
        return
    series, labels = read_twolead(ucr)

    window, psi = 10, 2
    dists = dtw.distance_matrix_fast(series, window=window, psi=psi)

    if directory:
        fn = directory / "matrix.png"
        dtwvis.plot_matrix(dists, filename=fn, shownumbers=True)
        print(f"Figure saved to {fn}")


if __name__ == "__main__":
    # Print options
    with util_numpy.test_uses_numpy() as np:
        np.set_printoptions(precision=2)
    # Logger options
    logger.setLevel(logging.DEBUG)
    sh = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('%(message)s')
    sh.setFormatter(formatter)
    logger.addHandler(sh)
    logger.propagate = 0
    # Output path
    directory = Path(__file__).resolve().parent.parent / "tests" / "output"

    ucr = Path("/Users/wannes/Projects/Datasets/UCR_TS_Archive_2015/")

    # test_distance_twolead_startdiff(directory=directory, ucr=ucr)
    # test_distance_twolead_sameclass(directory=directory, ucr=ucr)
    # test_distance_twolead_diffclass(directory=directory, ucr=ucr)
    test_distance_twolead_matrix(directory=directory, ucr=ucr)



# -*- coding: UTF-8 -*-
"""
dtaidistance.medoids
~~~~~~~~~~~~~~~~~~~~

Time series clustering.

:author: Wannes Meert
:copyright: Copyright 2020 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
import logging
import random
from pathlib import Path

try:
    import numpy as np
except ImportError:
    np = None

from ..util import SeriesContainer
from ..exceptions import NumpyException, PyClusteringException, MatplotlibException
from .visualization import prepare_plot_options


logger = logging.getLogger("be.kuleuven.dtai.distance")


class Medoids:
    def __init__(self, dists_fun, dists_options, k, show_progress=True):
        """

        :param dists_fun:
        :param dists_options:
        :param show_progress:
        """
        self.dists_fun = dists_fun
        self.dists_options = dists_options
        self.show_progress = show_progress
        self.k = k
        self.series = None
        self.cluster_idx = None

    def plot(self, filename=None, axes=None, ts_height=10,
             bottom_margin=2, top_margin=2, ts_left_margin=0, ts_sample_length=1,
             tr_label_margin=3, tr_left_margin=2, ts_label_margin=0,
             show_ts_label=None, show_tr_label=None,
             cmap='viridis_r', ts_color=None):

        try:
            from matplotlib import pyplot as plt
            from matplotlib.lines import Line2D
            import matplotlib.colors as colors
            import matplotlib.cm as cmx
        except ImportError:
            raise MatplotlibException("The plot function requires Matplotlib to be installed.")

        show_ts_label, show_tr_label = prepare_plot_options(show_ts_label, show_tr_label)
        self._series_y = [0] * len(self.series)
        max_y = self.series.get_max_y()
        self.ts_height_factor = (ts_height / max_y) * 0.9

        if axes is None:
            fig, ax = plt.subplots(nrows=1, ncols=2, frameon=False)
        else:
            fig, ax = None, axes
        ax[0].set_axis_off()
        ax[0].set_xlim(left=0, right=tr_left_margin + ts_sample_length * len(self.series[0]))
        ax[0].set_ylim(bottom=0, top=bottom_margin + ts_height * len(self.series) + top_margin)
        ax[1].set_axis_off()
        ax[1].set_xlim(left=0, right=ts_left_margin + ts_sample_length * len(self.series[0]))
        ax[1].set_ylim(bottom=0, top=bottom_margin + ts_height * len(self.series) + top_margin)

        if type(cmap) == str:
            cmap = plt.get_cmap(cmap)
        else:
            pass
        cluster_colors = cmx.ScalarMappable(norm=colors.Normalize(vmin=0, vmax=self.k), cmap=cmap)

        cnt_ts = 0
        for medoid_idx, medoid_id in enumerate(self.cluster_idx.keys()):
            for node in self.cluster_idx[medoid_id]:
                self._series_y[int(node)] = bottom_margin + ts_height * cnt_ts
                serie = self.series[int(node)]

                if ts_color:
                    curcolor = ts_color(int(node))
                else:
                    curcolor = cluster_colors.to_rgba(medoid_idx)

                if node == medoid_id:
                    ax[0].plot(ts_left_margin + ts_sample_length * np.arange(len(serie)),
                               bottom_margin + ts_height * cnt_ts + self.ts_height_factor * serie,
                               color=curcolor)

                ax[1].text(ts_left_margin + ts_label_margin,
                           bottom_margin + ts_height * cnt_ts + ts_height / 2,
                           show_ts_label(int(node)), ha='left', va='center')

                ax[1].plot(ts_left_margin + ts_sample_length * np.arange(len(serie)),
                           bottom_margin + ts_height * cnt_ts + self.ts_height_factor * serie,
                           color=curcolor)
                cnt_ts += 1

        if filename:
            if isinstance(filename, Path):
                filename = str(filename)
            plt.savefig(filename, bbox_inches='tight', pad_inches=0)
            plt.close()
            fig, ax = None, None

        return fig, ax


class KMedoids(Medoids):
    """KMedoids using the PyClustering package.

    Novikov, A., 2019. PyClustering: Data Mining Library. Journal of Open Source Software, 4(36), p.1230.
    Available at: http://dx.doi.org/10.21105/joss.01230.

    https://pyclustering.github.io/docs/0.9.0/html/d0/dd3/classpyclustering_1_1cluster_1_1kmedoids_1_1kmedoids.html

    """
    def __init__(self, dists_fun, dists_options, k=None, initial_medoids=None, show_progress=True):
        dists_options['compact'] = False
        self.initial_medoids = initial_medoids
        if k is None:
            if initial_medoids is None:
                raise AttributeError('Both k and initial_medoids cannot be None')
            k = len(initial_medoids)
        elif initial_medoids is not None and k != len(initial_medoids):
            raise AttributeError('The length of initial_medoids and k should be identical (or one of the two None)')
        super().__init__(dists_fun, dists_options, k, show_progress)

    def fit(self, series):
        try:
            from pyclustering.cluster.kmedoids import kmedoids
            from pyclustering.utils import calculate_distance_matrix
        except ImportError:
            raise PyClusteringException("The fit function requires the PyClustering package to be installed.")
        if np is None:
            raise NumpyException("The fit function requires Numpy to be installed.")
        self.series = SeriesContainer.wrap(series)
        logger.debug(f'KMedoid: Compute distances between {len(self.series)} series')
        dists = self.dists_fun(self.series, **self.dists_options)
        # Make the matrix symmetric
        i_lower = np.tril_indices(len(self.series), -1)
        dists[i_lower] = dists.T[i_lower]
        dists[np.isinf(dists)] = 0
        if self.initial_medoids is None:
            self.initial_medoids = random.sample(range(len(series)), k=self.k)
        kmedoids_instance = kmedoids(dists, self.initial_medoids, data_type='distance_matrix')
        kmedoids_instance.process()
        clusters = kmedoids_instance.get_clusters()
        medoids = kmedoids_instance.get_medoids()
        self.cluster_idx = {medoid: {inst for inst in instances} for medoid, instances in zip(medoids, clusters)}
        return self.cluster_idx



import os
import sys
import math
import tempfile
import pytest
import logging
from pathlib import Path

from dtaidistance import dtw, clustering, util_numpy
import dtaidistance.dtw_visualisation as dtwvis


directory = None
logger = logging.getLogger("be.kuleuven.dtai.distance")
numpyonly = pytest.mark.skipif("util_numpy.test_without_numpy()")
scipyonly = pytest.mark.skipif("util_numpy.test_without_scipy()")


@scipyonly
@numpyonly
def test_bug1():
    with util_numpy.test_uses_numpy() as np:
        series = np.array([
            [0., 0, 1, 2, 1, 0, 1, 0, 0],
            [0., 1, 2, 0, 0, 0, 0, 0, 0],
            [1., 2, 0, 0, 0, 0, 0, 1, 1],
            [0., 0, 1, 2, 1, 0, 1, 0, 0],
            [0., 1, 2, 0, 0, 0, 0, 0, 0],
            [1., 2, 0, 0, 0, 0, 0, 1, 1]])
        model = clustering.LinkageTree(dtw.distance_matrix_fast, {})
        cluster_idx = model.fit(series)

        if directory:
            hierarchy_fn = directory / "hierarchy.png"
        else:
            file = tempfile.NamedTemporaryFile()
            hierarchy_fn = Path(file.name + "_hierarchy.png")
        if not dtwvis.test_without_visualization():
            model.plot(hierarchy_fn)
            print("Figure saved to", hierarchy_fn)


@numpyonly
def test_bug2():
    with util_numpy.test_uses_numpy() as np:
        s1 = np.array([0, 0, 1, 2, 1, 0, 1, 0, 0], dtype=np.double)
        s2 = np.array([0.0, 1, 2, 0, 0, 0, 0, 0, 0])
        d1a = dtw.distance_fast(s1, s2, window=2)
        d1b = dtw.distance(s1, s2, window=2)

        if directory:
            fn = directory / "warpingpaths.png"
        else:
            file = tempfile.NamedTemporaryFile()
            fn = Path(file.name + "_warpingpaths.png")
        d2, paths = dtw.warping_paths(s1, s2, window=2)
        best_path = dtw.best_path(paths)
        if not dtwvis.test_without_visualization():
            dtwvis.plot_warpingpaths(s1, s2, paths, best_path, filename=fn, shownumbers=False)
            print("Figure saved to", fn)

        assert d1a == pytest.approx(d2)
        assert d1b == pytest.approx(d2)


@scipyonly
@numpyonly
def test_bug3():
    with util_numpy.test_uses_numpy() as np:
        series = np.array([
            np.array([1, 2, 1]),
            np.array([0., 1, 2, 0, 0, 0, 0, 0, 0]),
            np.array([1., 2, 0, 0, 0, 0, 0, 1, 1, 3, 4, 5]),
            np.array([0., 0, 1, 2, 1, 0, 1]),
            np.array([0., 1, 2, 0, 0, 0, 0, 0]),
            np.array([1., 2, 0, 0, 0, 0, 0, 1, 1])])
        ds = dtw.distance_matrix(series)
        print(ds)

        model = clustering.LinkageTree(dtw.distance_matrix, {})
        cluster_idx = model.fit(series)
        print(cluster_idx)

        if directory:
            fn = directory / "bug3.png"
        else:
            file = tempfile.NamedTemporaryFile()
            fn = Path(file.name + "_bug3.png")

        if not dtwvis.test_without_visualization():
            model.plot(fn, show_ts_label=True)


@numpyonly
def test_bug4():
    with util_numpy.test_uses_numpy() as np:
        s1 = np.array([0., 0, 1, 2, 1, 0, 1, 0, 0, 2, 1, 0, 0])
        s2 = np.array([0., 1, 2, 3, 1, 0, 0, 0, 2, 1, 0, 0, 0])
        path = dtw.warping_path(s1, s2)

        if directory:
            fn = directory / "bug4.png"
        else:
            file = tempfile.NamedTemporaryFile()
            fn = Path(file.name + "_bug4.png")

        if not dtwvis.test_without_visualization():
            dtwvis.plot_warping(s1, s2, path, filename=str(fn))


if __name__ == "__main__":
    logger.setLevel(logging.INFO)
    sh = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('%(message)s')
    sh.setFormatter(formatter)
    logger.addHandler(sh)
    logger.propagate = 0
    directory = Path(os.environ.get('TESTDIR', Path(__file__).parent))
    print(f"Saving files to {directory}")
    # test_bug1()
    test_bug2()
    # test_bug3()
    # test_bug4()



# -*- coding: UTF-8 -*-
"""
dtaidistance.util
~~~~~~~~~~~~~~~~~

Utility functions for DTAIDistance.

:author: Wannes Meert
:copyright: Copyright 2017-2018 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
import os
import sys
import csv
import logging
from array import array
from pathlib import Path
import tempfile


try:
    import numpy as np
except ImportError:
    np = None

try:
    from . import dtw_cc
except ImportError:
    dtw_cc = None

try:
    from . import dtw_cc_omp
except ImportError:
    dtw_cc_omp = None

try:
    from . import dtw_cc_numpy
except ImportError:
    dtw_cc_numpy = None


logger = logging.getLogger("be.kuleuven.dtai.distance")


dtaidistance_dir = os.path.abspath(os.path.dirname(__file__))


def prepare_directory(directory=None):
    """Prepare the given directory, create it if necessary.
    If no directory is given, a new directory will be created in the system's temp directory.
    """
    if directory is not None:
        directory = Path(directory)
        if not directory.exists():
            directory.mkdir(parents=True)
        logger.debug("Using directory: {}".format(directory))
        return Path(directory)
    directory = tempfile.mkdtemp(prefix="dtaidistance_")
    logger.debug("Using directory: {}".format(directory))
    return Path(directory)


def read_substitution_matrix(file):
    """Read substitution matrix from file.

    Comments starting with # and newlines are allowed anywhere
    in the file.
    
    :return: A dictionary mapping tuples of symbols to their weight.
    """

    def strip_comments(reader):
        for line in reader:
            if not line.rstrip() or line[0] == '#':
                continue
            yield line.rstrip()

    matrix = dict()
    with open(file) as f:
        reader = csv.reader(strip_comments(f), delimiter=" ", skipinitialspace=True)
        line = next(reader)
        idx = {i: symbol for i, symbol in enumerate(line)}
        for line in reader:
            symbol = line[0]
            for j, value in enumerate(line[1:]):
                matrix[(idx[j], symbol)] = float(value)
    return matrix


class SeriesContainer:
    def __init__(self, series, support_ndim=True):
        """Container for a list of series.

        This wrapper class knows how to deal with multiple types of datastructures to represent
        a list of sequences:
        - List[array.array]
        - List[numpy.array]
        - List[List]
        - numpy.array
        - numpy.matrix

        When using the C-based extensions, the data is automatically verified and converted.
        """
        self.support_ndim = support_ndim
        self.detected_ndim = False
        if isinstance(series, SeriesContainer):
            self.series = series.series
        elif np is not None and isinstance(series, np.ndarray):
            # A matrix always returns a 2D array, also if you select one row (to be consistent
            # and always be a matrix datastructure). The methods in this toolbox expect a
            # 1D array thus we need to convert to a 1D or 2D array. This is taken care by asarray
            self.series = np.asarray(series, order="C")
            if self.series.ndim > 2:
                if not self.support_ndim:
                    raise Exception(f'N-dimensional series are not supported '
                                    f'(series.ndim = {self.series.ndim})')
                else:
                    self.detected_ndim = True
        elif type(series) == set or type(series) == tuple or type(series) == list:
            self.series = list(series)
            if np is not None and isinstance(self.series[0], np.ndarray):
                if self.series[0].ndim > 1:
                    if not self.support_ndim:
                        raise Exception(f'N-dimensional series are not supported '
                                        f'(series[0].ndim = {self.series[0].ndim})')
                    else:
                        self.detected_ndim = True
        else:
            self.series = series

    def c_data_compat(self):
        """Return a datastructure that the C-component knows how to handle.
        The method tries to avoid copying or reallocating memory.

        :return: Either a list of buffers or a two-dimensional buffer. The
            buffers are guaranteed to be C-contiguous and can thus be used
            as regular pointer-based arrays in C.
        """
        if dtw_cc is None:
            raise Exception('C library not loaded')
        if type(self.series) == list:
            for i in range(len(self.series)):
                serie = self.series[i]
                if np is not None and isinstance(serie, np.ndarray):
                    if not self.support_ndim and serie.ndim != 1:
                        raise Exception(f'N-dimensional arrays are not supported (serie.ndim = {serie.ndim})')
                    if not serie.flags.c_contiguous:
                        serie = np.asarray(serie, order="C")
                        self.series[i] = serie
                elif isinstance(serie, array):
                    pass
                else:
                    raise Exception(
                        "Type of series not supported, "
                        "expected numpy.array or array.array but got {}".format(
                            type(serie)
                        )
                    )
            return dtw_cc.dtw_series_from_data(self.series)
        elif np is not None and isinstance(self.series, np.ndarray):
            if not self.series.flags.c_contiguous:
                logger.warning("Numpy array not C contiguous, copying data.")
                self.series = self.series.copy(order="C")
            if not self.support_ndim and self.series.ndim > 2:
                raise Exception(f'N-dimensional series are not supported (series.ndim = {self.series.ndim})')
            if dtw_cc_numpy is None:
                logger.warning("DTAIDistance C-extension for Numpy is not available. Proceeding anyway.")
                return dtw_cc.dtw_series_from_data(self.series)
            elif len(self.series.shape) == 3:
                return dtw_cc_numpy.dtw_series_from_numpy_ndim(self.series)
            else:
                return dtw_cc_numpy.dtw_series_from_numpy(self.series)
        return dtw_cc.dtw_series_from_data(self.series)

    def get_max_y(self):
        max_y = 0
        if isinstance(self.series, np.ndarray) and len(self.series.shape) == 2:
            max_y = max(np.max(self.series), abs(np.min(self.series)))
        else:
            for serie in self.series:
                max_y = max(max_y, np.max(serie), abs(np.min(serie)))
        return max_y

    def get_max_length(self):
        max_length = 0
        if isinstance(self.series, np.ndarray) and len(self.series.shape) == 2:
            max_length = self.series.shape[1]
        else:
            for serie in self.series:
                max_length = max(max_length, len(serie))
        return max_length

    def get_avg_length(self):
        max_length = 0
        if isinstance(self.series, np.ndarray) and len(self.series.shape) == 2:
            max_length = self.series.shape[1]
        else:
            for serie in self.series:
                max_length += len(serie)
            max_length /= len(self.series)
        return max_length

    def __getitem__(self, item):
        return self.series[item]

    def __len__(self):
        return len(self.series)

    def __str__(self):
        return "SeriesContainer:\n{}".format(self.series)

    @staticmethod
    def wrap(series, support_ndim=True):
        if isinstance(series, SeriesContainer):
            series.support_ndim = support_ndim
            return series
        return SeriesContainer(series, support_ndim=support_ndim)


def recompile():
    import subprocess as sp

    sp.run([sys.executable, "setup.py", "build_ext", "--inplace"], cwd=dtaidistance_dir)


def argmin(a):
    imin, vmin = 0, float("inf")
    for i, v in enumerate(a):
        if v < vmin:
            imin, vmin = i, v
    return imin



# -*- coding: UTF-8 -*-
"""
dtaidistance.dtw_visualisation
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Dynamic Time Warping (DTW) visualisations.

:author: Wannes Meert
:copyright: Copyright 2017 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
import os
import logging

from . import util_numpy


try:
    if util_numpy.test_without_numpy():
        raise ImportError()
    import numpy as np
except ImportError:
    np = None


logger = logging.getLogger("be.kuleuven.dtai.distance")

from . import dtw
try:
    from . import dtw_c
except ImportError:
    # logger.info('C library not available')
    dtw_c = None

try:
    from tqdm import tqdm
except ImportError:
    logger.info('tqdm library not available')
    tqdm = None


def test_without_visualization():
    if "DTAIDISTANCE_TESTWITHOUTVIZ" in os.environ and os.environ["DTAIDISTANCE_TESTWITHOUTVIZ"] == "1":
        return True
    return False


def plot_warp(from_s, to_s, new_s, path, filename=None):
    """Plot the warped sequence and its relation to the original sequence
    and the target sequence.

    :param from_s: From sequence.
    :param to_s: To sequence.
    :param new_s: Warped version of from sequence.
    :param path: Optimal warping path.
    :param filename: Filename path (optional).
    """
    try:
        import matplotlib.pyplot as plt
        import matplotlib as mpl
        from matplotlib.patches import ConnectionPatch
    except ImportError:
        logger.error("The plot_warp function requires the matplotlib package to be installed.")
        return
    fig, ax = plt.subplots(nrows=3, ncols=1, sharex='all', sharey='all')
    ax[0].plot(from_s, label="From")
    ax[0].legend()
    ax[1].plot(to_s, label="To")
    ax[1].legend()
    lines = []
    line_options = {'linewidth': 0.5, 'color': 'orange', 'alpha': 0.8}
    for r_c, c_c in path:
        if r_c < 0 or c_c < 0:
            continue
        con = ConnectionPatch(xyA=[r_c, from_s[r_c]], coordsA=ax[0].transData,
                              xyB=[c_c, to_s[c_c]], coordsB=ax[1].transData, **line_options)

        lines.append(con)
    ax[2].plot(new_s, label="From-warped")
    ax[2].legend()
    for i in range(len(to_s)):
        con = ConnectionPatch(xyA=[i, to_s[i]], coordsA=ax[1].transData,
                              xyB=[i, new_s[i]], coordsB=ax[2].transData, **line_options)
        lines.append(con)
    for line in lines:
        fig.add_artist(line)
    if filename:
        plt.savefig(filename)
        plt.close()
        fig, ax = None, None
    return fig, ax


def plot_warping(s1, s2, path, filename=None):
    """Plot the optimal warping between to sequences.

    :param s1: From sequence.
    :param s2: To sequence.
    :param path: Optimal warping path.
    :param filename: Filename path (optional).
    """
    try:
        import matplotlib.pyplot as plt
        import matplotlib as mpl
        from matplotlib.patches import ConnectionPatch
    except ImportError:
        logger.error("The plot_warp function requires the matplotlib package to be installed.")
        return
    fig, ax = plt.subplots(nrows=2, ncols=1, sharex='all', sharey='all')
    ax[0].plot(s1)
    ax[1].plot(s2)
    plt.tight_layout()
    lines = []
    line_options = {'linewidth': 0.5, 'color': 'orange', 'alpha': 0.8}
    for r_c, c_c in path:
        if r_c < 0 or c_c < 0:
            continue
        con = ConnectionPatch(xyA=[r_c, s1[r_c]], coordsA=ax[0].transData,
                              xyB=[c_c, s2[c_c]], coordsB=ax[1].transData, **line_options)
        lines.append(con)
    for line in lines:
        fig.add_artist(line)
    if filename:
        plt.savefig(filename)
        plt.close()
        fig, ax = None, None
    return fig, ax


def plot_warpingpaths(s1, s2, paths, path=None, filename=None, shownumbers=False, showlegend=False):
    """Plot the warping paths matrix.

    :param s1: Series 1
    :param s2: Series 2
    :param paths: Warping paths matrix
    :param path: Path to draw (typically this is the best path)
    :param filename: Filename for the image (optional)
    :param shownumbers: Show distances also as numbers
    :param showlegend: Show colormap legend
    """
    try:
        from matplotlib import pyplot as plt
        from matplotlib import gridspec
        from matplotlib.ticker import FuncFormatter
    except ImportError:
        logger.error("The plot_warpingpaths function requires the matplotlib package to be installed.")
        return
    ratio = max(len(s1), len(s2))
    min_y = min(np.min(s1), np.min(s2))
    max_y = max(np.max(s1), np.max(s2))

    fig = plt.figure(figsize=(10, 10), frameon=True)
    if showlegend:
        grows = 3
        gcols = 3
        height_ratios = [1, 6, 1]
        width_ratios = [1, 6, 1]
    else:
        grows = 2
        gcols = 2
        height_ratios = [1, 6]
        width_ratios = [1, 6]
    gs = gridspec.GridSpec(grows, gcols, wspace=1, hspace=1,
                           left=0, right=10.0, bottom=0, top=1.0,
                           height_ratios=height_ratios,
                           width_ratios=width_ratios)
    max_s2_x = np.max(s2)
    max_s2_y = len(s2)
    max_s1_x = np.max(s1)
    min_s1_x = np.min(s1)
    max_s1_y = len(s1)

    if path is None:
        p = dtw.best_path(paths)
    else:
        p = path

    def format_fn2_x(tick_val, tick_pos):
        return max_s2_x - tick_val

    def format_fn2_y(tick_val, tick_pos):
        return int(max_s2_y - tick_val)

    ax0 = fig.add_subplot(gs[0, 0])
    ax0.set_axis_off()
    ax0.text(0, 0, "Dist = {:.4f}".format(paths[p[-1][0], p[-1][1]]))
    ax0.xaxis.set_major_locator(plt.NullLocator())
    ax0.yaxis.set_major_locator(plt.NullLocator())

    ax1 = fig.add_subplot(gs[0, 1])
    ax1.set_ylim([min_y, max_y])
    ax1.set_axis_off()
    ax1.xaxis.tick_top()
    # ax1.set_aspect(0.454)
    ax1.plot(range(len(s2)), s2, ".-")
    ax1.xaxis.set_major_locator(plt.NullLocator())
    ax1.yaxis.set_major_locator(plt.NullLocator())

    ax2 = fig.add_subplot(gs[1, 0])
    ax2.set_xlim([-max_y, -min_y])
    ax2.set_axis_off()
    # ax2.set_aspect(0.8)
    # ax2.xaxis.set_major_formatter(FuncFormatter(format_fn2_x))
    # ax2.yaxis.set_major_formatter(FuncFormatter(format_fn2_y))
    ax2.xaxis.set_major_locator(plt.NullLocator())
    ax2.yaxis.set_major_locator(plt.NullLocator())
    ax2.plot(-s1, range(max_s1_y, 0, -1), ".-")

    ax3 = fig.add_subplot(gs[1, 1])
    # ax3.set_aspect(1)
    img = ax3.matshow(paths[1:, 1:])
    # ax3.grid(which='major', color='w', linestyle='-', linewidth=0)
    # ax3.set_axis_off()
    py, px = zip(*p)
    ax3.plot(px, py, ".-", color="red")
    # ax3.xaxis.set_major_locator(plt.NullLocator())
    # ax3.yaxis.set_major_locator(plt.NullLocator())
    if shownumbers:
        for r in range(1, paths.shape[0]):
            for c in range(1, paths.shape[1]):
                ax3.text(c - 1, r - 1, "{:.2f}".format(paths[r, c]))

    gs.tight_layout(fig, pad=1.0, h_pad=1.0, w_pad=1.0)
    # fig.subplots_adjust(hspace=0, wspace=0)

    if showlegend:
        # ax4 = fig.add_subplot(gs[0:, 2])
        ax4 = fig.add_axes([0.9, 0.25, 0.015, 0.5])
        fig.colorbar(img, cax=ax4)

    ax = fig.axes

    if filename:
        if type(filename) != str:
            filename = str(filename)
        plt.savefig(filename)
        plt.close()
        fig, ax = None, None
    return fig, ax

def plot_matrix(distances, filename=None, ax=None, shownumbers=False):
    from matplotlib import pyplot as plt

    if ax is None:
        if shownumbers:
            figsize = (15, 15)
        else:
            figsize = None
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=figsize)
    else:
        fig = None

    ax.xaxis.set_ticks_position('top')
    ax.yaxis.set_ticks_position('both')

    im = ax.imshow(distances)
    idxs = [str(i) for i in range(len(distances))]
    # Show all ticks
    ax.set_xticks(np.arange(len(idxs)))
    ax.set_xticklabels(idxs)
    ax.set_yticks(np.arange(len(idxs)))
    ax.set_yticklabels(idxs)

    ax.set_title("Distances between series", pad=30)

    if shownumbers:
        for i in range(len(idxs)):
            for j in range(len(idxs)):
                if not np.isinf(distances[i, j]):
                    l = "{:.2f}".format(distances[i, j])
                    ax.text(j, i, l, ha="center", va="center", color="w")

    if filename:
        if type(filename) != str:
            filename = str(filename)
        plt.savefig(filename)
        plt.close()
        fig, ax = None, None
    return fig, ax

def plot_average(s1, s2, avg, path1, path2, filename=None, ax=None):
    """Plot how s1 and s2 relate to the avg.

    :param s1: Seq 1.
    :param s2: Seq 2.
    :param path: Average sequence.
    :param filename: Filename path (optional).
    """
    try:
        import matplotlib.pyplot as plt
        import matplotlib as mpl
        from matplotlib.patches import ConnectionPatch
    except ImportError:
        logger.error("The plot_warp function requires the matplotlib package to be installed.")
        return
    if ax is None:
        fig, ax = plt.subplots(nrows=1, ncols=1, sharex='all', sharey='all')
    else:
        fig = None
    ax.plot(s1, color='blue', alpha=0.5)
    ax.plot(s2, color='blue', alpha=0.5)
    ax.plot(avg, color='orange', linestyle='dashed', alpha=0.5)
    # plt.tight_layout()
    # lines = []
    # line_options = {'linewidth': 0.5, 'color': 'orange', 'alpha': 0.8}
    # for r_c, c_c in path:
    #     if r_c < 0 or c_c < 0:
    #         continue
    #     con = ConnectionPatch(xyA=[r_c, s1[r_c]], coordsA=ax[0].transData,
    #                           xyB=[c_c, s2[c_c]], coordsB=ax[1].transData, **line_options)
    #     lines.append(con)
    # for line in lines:
    #     fig.add_artist(line)
    if filename:
        plt.savefig(filename)
        plt.close()
        fig, ax = None, None
    return fig, ax


import os
from pathlib import Path

import pytest

from dtaidistance import alignment, util_numpy
from dtaidistance.util import read_substitution_matrix


numpyonly = pytest.mark.skipif("util_numpy.test_without_numpy()")
directory = Path()


@numpyonly
def test_sequences1():
    """Example from https://en.wikipedia.org/wiki/Needleman–Wunsch_algorithm . """
    with util_numpy.test_uses_numpy() as np:
        s1 = "GATTACA"
        s2 = "GCATGCU"
        value, matrix = alignment.needleman_wunsch(s1, s2)
        algn, s1a1, s2a1 = alignment.best_alignment(matrix, s1, s2, gap='-')
        matrix_sol = [
            [-0., -1., -2., -3., -4., -5., -6., -7.],
            [-1.,  1., -0., -1., -2., -3., -4., -5.],
            [-2., -0., -0.,  1., -0., -1., -2., -3.],
            [-3., -1., -1., -0.,  2.,  1., -0., -1.],
            [-4., -2., -2., -1.,  1.,  1., -0., -1.],
            [-5., -3., -3., -1., -0., -0., -0., -1.],
            [-6., -4., -2., -2., -1., -1.,  1., -0.],
            [-7., -5., -3., -1., -2., -2., -0., -0.]]
        algn_sol1 = [['G', '-', 'A', 'T', 'T', 'A', 'C', 'A'], ['G', 'C', 'A', 'T', '-', 'G', 'C', 'U']]
        assert value == 0.0
        assert np.array_equal(matrix, matrix_sol)
        assert s1a1 == algn_sol1[0]
        assert s2a1 == algn_sol1[1]


def test_sequences2():
    s1 = "GAAAAAAAT"
    s2 = "GAAT"
    value, matrix = alignment.needleman_wunsch(s1, s2)
    algn, s1a1, s2a1 = alignment.best_alignment(matrix, s1, s2, gap='-')
    algn_sol1 = [list('GAAAAAAAT'), list('GAA-----T')]
    assert s1a1 == algn_sol1[0]
    assert s2a1 == algn_sol1[1]


def test_sequences3():
    s1 = "GAAAAAAAT"
    s2 = "GAATA"
    value, matrix = alignment.needleman_wunsch(s1, s2)
    algn, s1a1, s2a1 = alignment.best_alignment(matrix, s1, s2, gap='-')
    algn_sol1 = [list('GAA-AAAAAT'), list('GAATA-----')]
    assert s1a1 == algn_sol1[0]
    assert s2a1 == algn_sol1[1]


def test_sequences4():
    s1 = "AGACTAGTTACC"
    s2 = "CGAGACGTC"
    value, matrix = alignment.needleman_wunsch(s1, s2)
    algn, s1a1, s2a1 = alignment.best_alignment(matrix, s1, s2, gap='-')
    # print(matrix)
    # print(algn)
    # print(s1a1)
    # print(s2a1)
    algn_sol1 = [list("--AGACTAGTTACC"), list("CGAGAC--GT--C-")]
    assert s1a1 == algn_sol1[0]
    assert s2a1 == algn_sol1[1]


def test_sequences_custom():

    matrix = read_substitution_matrix(
        Path(__file__).parent / "rsrc" / "substitution.txt"
    )
    substitution = alignment.make_substitution_fn(matrix)

    s1 = "CCAGG"
    s2 = "CCGA"

    value, matrix = alignment.needleman_wunsch(s1, s2)
    algn, s1a1, s2a1 = alignment.best_alignment(matrix, s1, s2, gap='-')
    algn_sol1 = [list("CCAGG"), list("CC-GA")]
    assert s1a1 == algn_sol1[0]
    assert s2a1 == algn_sol1[1]

    value, matrix = alignment.needleman_wunsch(s1, s2, substitution=substitution)
    algn, s1a2, s2a2 = alignment.best_alignment(matrix, s1, s2, gap='-')
    algn_sol2 = [list("CC-AGG"), list("CCGA--")]
    assert s1a2 == algn_sol2[0]
    assert s2a2 == algn_sol2[1]


def test_sequences_blosum():
    matrix = read_substitution_matrix(
        Path(__file__).parent / "rsrc" / "substitution.txt"
    )
    substitution = alignment.make_substitution_fn(matrix)
    s1 = "AGACTAGTTAC"
    s2 = "CGAGACGT"
    value, matrix = alignment.needleman_wunsch(s1, s2, substitution=substitution)
    algn, s1a1, s2a1 = alignment.best_alignment(matrix, s1, s2, gap='-')
    algn_sol1 = [list('--AGACTAGTTAC'),
                 list('CGAGAC--GT---')]
    assert s1a1 == algn_sol1[0]
    assert s2a1 == algn_sol1[1]               


def test_substitution_function():
    matrix = read_substitution_matrix(
        Path(__file__).parent / "rsrc" / "substitution.txt"
    )
    substitution = alignment.make_substitution_fn(matrix, gap=0.5)
    assert substitution('A', 'B') == (1, 0.5)
    assert substitution('A', 'A') == (-20.0, 0.5)
    assert substitution('A', 'C') == (3.0, 0.5)

    # test just changing gap function
    substitution2 = alignment.make_substitution_fn({}, gap=0.5)
    assert substitution2('A', 'A') == (-1, 0.5)
    assert substitution2('G', 'C') == (1, 0.5)
    assert alignment._default_substitution_fn('A', 'A') == (-1, 1)
    assert alignment._default_substitution_fn('G', 'C') == (1, 1)


if __name__ == "__main__":
    directory = Path(os.environ.get('TESTDIR', Path(__file__).parent))
    print(f"Saving files to {directory}")
    test_sequences1()
    test_sequences2()
    test_sequences3()
    test_sequences4()
    test_sequences_custom()
    test_sequences_blosum()
    test_substitution_function()



cimport dtaidistancec_dtw

cdef class DTWBlock:
    cdef dtaidistancec_dtw.DTWBlock _block

cdef class DTWSettings:
    cdef dtaidistancec_dtw.DTWSettings _settings

cdef class DTWSeriesPointers:
    cdef double **_ptrs
    cdef Py_ssize_t *_lengths
    cdef Py_ssize_t _nb_ptrs

cdef class DTWSeriesMatrix:
    cdef double[:,::1] _data

cdef class DTWSeriesMatrixNDim:
    cdef double[:,:,::1] _data



# -*- coding: UTF-8 -*-
"""
dtaidistance.dtw_barycenter
~~~~~~~~~~~~~~~~~~~~~~~~~~~

Dynamic Time Warping (DTW) Barycenter

:author: Wannes Meert
:copyright: Copyright 2020 KU Leuven, DTAI Research Group.
:license: Apache License, Version 2.0, see LICENSE for details.

"""
import logging
import math
import array
import random

from .dtw import warping_path, distance_matrix
from . import ed
from . import util
from . import util_numpy
from .util import SeriesContainer
from .exceptions import NumpyException


logger = logging.getLogger("be.kuleuven.dtai.distance")


dtw_cc = None
try:
    from . import dtw_cc
except ImportError:
    logger.debug('DTAIDistance C library not available')
    dtw_cc = None


try:
    if util_numpy.test_without_numpy():
        raise ImportError()
    import numpy as np
except ImportError:
    np = None


def get_good_c(s, mask, nb_initial_samples, use_c=False, **kwargs):
    if nb_initial_samples > len(s):
        nb_initial_samples = len(s)
    mask_size = np.sum(mask)
    cs = []
    randthr = nb_initial_samples / mask_size
    for i in range(len(s)):
        if mask[i]:
            if random.random() <= randthr:
                cs.append(s[i])
        if len(cs) == nb_initial_samples:
            break
        else:
            randthr = (nb_initial_samples - len(cs)) / (mask_size - i - 1)
    d = distance_matrix(cs, use_c=use_c,  **kwargs)
    d = d.sum(axis=1)
    best_i = np.argmin(d)
    return s[best_i]


def dba_loop(s, c=None, max_it=10, thr=0.001, mask=None,
             keep_averages=False, use_c=False, nb_initial_samples=None, **kwargs):
    """Loop around the DTW Barycenter Averaging (DBA) method until convergence.

    :param s: Container of sequences
    :param c: Initial averaging sequence.
        If none is given, the first one is used (unless if nb_initial_samples is set).
        Better performance can be achieved by starting from an informed
        starting point (Petitjean et al. 2011).
    :param max_it: Maximal number of calls to DBA.
    :param thr: Convergence if the DBA is changing less than this value.
    :param mask: Boolean array with the series in s to use. If None, use all.
    :param keep_averages: Keep all DBA values (for visualisation or debugging).
    :param nb_initial_samples: If c is None, and this argument is not None, select
        nb_initial_samples samples and select the series closest to all other samples
        as c.
    :param use_c: Use a fast C implementation instead of a Python version.
    :param kwargs: Arguments for dtw.distance
    """
    if np is None:
        raise NumpyException('The method dba_loop requires Numpy to be available')
    s = SeriesContainer.wrap(s)
    avg = None
    avgs = None
    if keep_averages:
        avgs = []
    if mask is None:
        mask = np.full((len(s),), True, dtype=bool)
    if c is None:
        if nb_initial_samples is None:
            curi = 0
            while mask[curi] is False:
                curi += 1
            c = s[curi]
        else:
            c = get_good_c(s, mask, nb_initial_samples, use_c=use_c, **kwargs)

        # You can also use a constant function, but this gives worse performance.
        # After the first iteration, this will be the average of all
        # sequences. The disadvantage is that this might create e.g. multiple
        # peaks for a sequence with only one peak (but shifted) and then the
        # original sequences will map their single peak to the different peaks
        # in the first average and converge to that as a local optimum.
        # t = s.get_avg_length()
        # c = array.array('d', [0] * t)
    for it in range(max_it):
        logger.debug(f'DBA Iteration {it}')
        if use_c:
            assert(c is not None)
            c_copy = c.copy()  # The C code reuses this array
            if np is not None and isinstance(mask, np.ndarray):
                # The C code requires a bit array of uint8 (or unsigned char)
                mask_copy = np.packbits(mask, bitorder='little')
            else:
                raise Exception('Mask only implemented for C when passing a Numpy array. '
                                f'Got {type(mask)}')
            dtw_cc.dba(s, c_copy, mask=mask_copy, **kwargs)
            avg = c_copy
        else:
            avg = dba(s, c, mask=mask, use_c=use_c, **kwargs)
        if keep_averages:
            avgs.append(avg)
        if thr is not None and c is not None:
            diff = 0
            # diff = np.sum(np.subtract(avg, c))
            for av, cv in zip(avg, c):
                diff += abs(av - cv)
            diff /= len(avg)
            if diff <= thr:
                logger.debug(f'DBA converged at {it} iterations (avg diff={diff}).')
                break
        c = avg
    if keep_averages:
        return avg, avgs
    return avg


def dba(s, c, mask=None, use_c=False, nb_initial_samples=None, **kwargs):
    """DTW Barycenter Averaging.

    F. Petitjean, A. Ketterlin, and P. Gan ̧carski.
    A global averaging method for dynamic time warping, with applications to clustering.
    Pattern Recognition, 44(3):678–693, 2011.

    :param s: Container of sequences
    :param c: Initial averaging sequence.
        If none is given, the first one is used (unless if nb_initial_samples is set).
        Better performance can be achieved by starting from an informed
        starting point (Petitjean et al. 2011).
    :param mask: Boolean array with the series in s to use. If None, use all.
    :param nb_initial_samples: If c is None, and this argument is not None, select
        nb_initial_samples samples and select the series closest to all other samples
        as c.
    :param use_c: Use a fast C implementation instead of a Python version.
    :param kwargs: Arguments for dtw.distance
    :return: Bary-center of length len(c).
    """
    s = SeriesContainer.wrap(s)
    if mask is not None and not mask.any():
        # Mask has not selected any series
        print("Empty mask, returning zero-constant average")
        c = array.array('d', [0] * len(s[0]))
        return c
    if mask is None:
        mask = np.full((len(s),), True, dtype=bool)
    if c is None:
        if nb_initial_samples is None:
            curi = 0
            while mask[curi] is False:
                curi += 1
            c = s[curi]
        else:
            c = get_good_c(s, mask, nb_initial_samples, use_c=use_c, **kwargs)
    t = len(c)
    assoctab = [[] for _ in range(t)]
    for idx, seq in enumerate(s):
        if mask is not None and not mask[idx]:
            continue
        if use_c:
            m = dtw_cc.warping_path(c, seq, **kwargs)
        else:
            m = warping_path(c, seq, **kwargs)
        for i, j in m:
            assoctab[i].append(seq[j])
    cp = array.array('d', [0] * t)
    for i, values in enumerate(assoctab):
        if len(values) == 0:
            print('WARNING: zero values in assoctab')
            print(c)
            for seq in s:
                print(seq)
            print(assoctab)
        cp[i] = sum(values) / len(values)  # barycenter
    return cp





from ..exceptions import MatplotlibException


def prepare_plot_options(show_ts_label, show_tr_label):
    if show_ts_label is True:
        show_ts_label = lambda idx: str(int(idx))
    elif show_ts_label is False or show_ts_label is None:
        show_ts_label = lambda idx: ""
    elif callable(show_ts_label):
        pass
    elif hasattr(show_ts_label, "__getitem__"):
        show_ts_label_prev = show_ts_label
        show_ts_label = lambda idx: show_ts_label_prev[idx]
    else:
        raise AttributeError("Unknown type for show_ts_label, expecting boolean, subscriptable or callable, "
                             "got {}".format(type(show_ts_label)))

    if show_tr_label is True:
        show_tr_label = lambda dist: "{:.2f}".format(dist)
    elif show_tr_label is False or show_tr_label is None:
        show_tr_label = lambda dist: ""
    elif callable(show_tr_label):
        pass
    elif hasattr(show_tr_label, "__getitem__"):
        show_tr_label_prev = show_tr_label
        show_tr_label = lambda idx: show_tr_label_prev[idx]
    else:
        raise AttributeError("Unknown type for show_ts_label, expecting boolean, subscriptable or callable, "
                             "got {}".format(type(show_ts_label)))

    return show_ts_label, show_tr_label
